/*
Copyright 2023 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package main

import (
	"bytes"
	"encoding/json"
	"fmt"
	"net/http"

	"github.com/google/go-containerregistry/pkg/name"
	v1 "github.com/google/go-containerregistry/pkg/v1"
	"github.com/google/go-containerregistry/pkg/v1/remote"
)

// layersForV1 gets the layers for a v1 schema image
func layersForV1(transport http.RoundTripper, ref name.Reference, desc *remote.Descriptor) ([]v1.Layer, error) {
	m := &schema1{}
	if err := json.NewDecoder(bytes.NewReader(desc.Manifest)).Decode(m); err != nil {
		return nil, err
	}
	layers := make([]v1.Layer, len(m.FSLayers))
	for i, fsLayer := range m.FSLayers {
		layerDigest, err := name.NewDigest(fmt.Sprintf("%s@%s", ref.Context(), fsLayer.BlobSum))
		if err != nil {
			return nil, err
		}
		layer, err := remote.Layer(layerDigest, remote.WithTransport(transport))
		if err != nil {
			return nil, err
		}
		layers[i] = layer
	}
	return layers, nil
}

type fslayer struct {
	BlobSum string `json:"blobSum"`
}

type schema1 struct {
	FSLayers []fslayer `json:"fsLayers"`
}
