/*
Copyright 2022 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package clientip

import (
	"net/http"
	"net/netip"
	"testing"
)

func TestGet(t *testing.T) {
	testCases := []struct {
		Name        string
		Request     http.Request
		ExpectedIP  netip.Addr
		ExpectError bool
	}{
		{
			Name: "NO X-Forwarded-For",
			Request: http.Request{
				RemoteAddr: "127.0.0.1:8888",
			},
			ExpectedIP: netip.MustParseAddr("127.0.0.1"),
		},
		{
			Name: "NO X-Forwarded-For, somehow bogus RemoteAddr ??? gotta pump code coverage 🤷",
			Request: http.Request{
				RemoteAddr: "127.0.0.1asd;lfkj8888",
			},
			ExpectError: true,
		},
		{
			Name: "X-Forwarded-For without client-supplied",
			Request: http.Request{
				Header: http.Header{
					"X-Forwarded-For": []string{"8.8.8.8,8.8.8.9"},
				},
				RemoteAddr: "127.0.0.1:8888",
			},
			ExpectedIP: netip.MustParseAddr("8.8.8.8"),
		},
		{
			Name: "X-Forwarded-For with clean client-supplied",
			Request: http.Request{
				Header: http.Header{
					"X-Forwarded-For": []string{"127.0.0.1, 8.8.8.8, 8.8.8.9"},
				},
				RemoteAddr: "127.0.0.1:8888",
			},
			ExpectedIP: netip.MustParseAddr("8.8.8.8"),
		},
		{
			Name: "X-Forwarded-For with garbage client-supplied",
			Request: http.Request{
				Header: http.Header{
					"X-Forwarded-For": []string{"asd;lfkjaasdf;lk,,8.8.8.8,8.8.8.9"},
				},
				RemoteAddr: "127.0.0.1:8888",
			},
			ExpectedIP: netip.MustParseAddr("8.8.8.8"),
		},
		{
			Name: "Bogus crafted non-cloud X-Forwarded-For with no commas",
			Request: http.Request{
				Header: http.Header{
					"X-Forwarded-For": []string{"8.8.8.8"},
				},
				RemoteAddr: "127.0.0.1:8888",
			},
			ExpectError: true,
		},
		{
			Name: "X-Forwarded-For for IPv6 with load balancer",
			Request: http.Request{
				Header: http.Header{
					"X-Forwarded-For": []string{"2001:0db8:1234:5678:abcd:1234:5678:abcd, 2001:0db8:0:abcd::"},
				},
				RemoteAddr: "127.0.0.1:8888",
			},
			ExpectedIP: netip.MustParseAddr("2001:0db8:1234:5678:abcd:1234:5678:abcd"),
		},
		{
			Name: "X-Forwarded-For for IPv6 without load balancer",
			Request: http.Request{
				Header: http.Header{
					"X-Forwarded-For": []string{"2001:0db8:1234:5678:abcd:1234:5678:abcd"},
				},
				RemoteAddr: "127.0.0.1:8888",
			},
			// We could accept this, we choose to require the load balancer
			ExpectError: true,
		},
	}
	for i := range testCases {
		tc := testCases[i]
		t.Run(tc.Name, func(t *testing.T) {
			//t.Parallel()
			ip, err := Get(&tc.Request)
			if err != nil {
				if !tc.ExpectError {
					t.Fatalf("unexpected error: %v", err)
				}
			} else if tc.ExpectError {
				t.Fatal("expected error but err was nil")
			} else if ip != tc.ExpectedIP {
				t.Fatalf("IP does not match expected IP got: %q, expected: %q", ip, tc.ExpectedIP)
			}
		})
	}
}
